<div style="padding-bottom: 3em">
  <form class="navbar-form pull-right" action="{{ pathto('search') }}" method="get">
    <input type="text" class="search span3" name="q" placeholder="Search documentation ...">
    <input type="hidden" name="check_keywords" value="yes" >
    <input type="hidden" name="area" value="default" >
  </form>
</div>
