<li><a href="/docs/stable/install.html">Installation</a></li>
<li><a href="{{ pathto('auto_examples/index') }}">Gallery</a></li>
<li><a href="{{ pathto('index') }}">Documentation</a></li>
<li><a href="/community_guidelines.html">Community</a></li>

<li><a href="https://github.com/scikit-image/scikit-image">
    <img src="{{ pathto('_static', 1) }}/GitHub-Mark-32px.png"
        style="height: 15px; width: 15px;
               display: inline; float: none;
               padding-bottom: 3px;">
    Source</a>
</li>
